/**
  *  \file client/dialogs/inboxdlg.cc
  *  \brief Incoming Message Dialog
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/inboxdlg.h"
#include "game/inbox.h"
#include "game/global.h"
#include "util/nls.h"
#include "ui/window.h"
#include "ui/layout.h"
#include "client/widgets/msgaction.h"
#include "gfx/palette.h"
#include "client/widgets/msgdisplay.h"
#include "ui/prefix.h"
#include "game/parser.h"
#include "game/msgparse.h"
#include "game/ship.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/msgglobal.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"
#include "client/widgets/helpwidget.h"

namespace {
    class WInboxActionPanel : public WMessageActionPanel {
        GInbox& inbox;
     public:
        WInboxActionPanel(GInbox& inbox)
            : WMessageActionPanel(inbox),
              inbox(inbox)
            {
                onBrowse();
            }
        void updateMessageControls();
    };

    class WInboxWindow : public UIWindow {
        ObjectHolder h;

        void init(GInbox& inbox);
     public:
        WInboxWindow(GInbox& inbox);
    };
}

void
WInboxActionPanel::updateMessageControls()
{
    // Split into lines
    GMessageLines lines;
    splitMessage(lines, inbox.getText(inbox.getCurrent()));

    // Check header for reply information
    int kind = getMessageHeaderInformation(lines, MsgHdrKind);
    int subid = getMessageHeaderInformation(lines, MsgHdrSubId);
    if (kind == 'r') {
        int n = getPlayerIdFromChar(subid);
        if (n >= 0 && n <= NUM_PLAYERS) {
            setHaveReply(true, n);
        }
    }
    if (kind == 'g' || kind == 'h') {
        setHaveReply(true, 12);
    }

    // Check content for object type
    ptr_vector<GMessageInformation> info;
    getGlobalMessageParser().parseMessage(inbox.getText(inbox.getCurrent()),
                                          inbox.getReceiver(inbox.getCurrent()),
                                          getCurrentTurn().getTurnNumber(),
                                          info);
    int q = 0;
    for (ptr_vector<GMessageInformation>::size_type i = 0; i < info.size(); ++i) {
        /* If a message talks about multiple objects, preferences are:
           - foreign ships (might have hit a mine or blown up)
           - minefields (might have been laid)
           - planets (might have been colonized)
           - own ships
           - ion storms */
        int thisq = 0;
        WScreenId thisobj;
        int thisid = info[i]->getObjectId();
        switch (info[i]->getObjectType()) {
         case MessageShip:
            if (global_anyship_type.isExistant(thisid)) {
                if (global_ship_type.isExistant(thisid)) {
                    thisobj = WScreenId(WScreenId::ShipScreen, thisid);
                    thisq = 20;
                } else {
                    thisobj = WScreenId(global_anyship_type.getObject(thisid).getPos());
                    thisq = 50;
                }
            }
            break;
         case MessageMinefield:
            if (global_minefield_type.isExistant(thisid)) {
                thisobj = WScreenId(global_minefield_type.getObject(thisid).getPos());
                thisq = 40;
            }
            break;
         case MessagePlanet:
         case MessageStarbase:
            if (global_anyplanet_type.isExistant(thisid)) {
                if (info[i]->getObjectType() == MessageStarbase && global_base_type.isExistant(thisid)) {
                    thisobj = WScreenId(WScreenId::BaseScreen, thisid);
                    thisq = 32;
                } else if (global_planet_type.isExistant(thisid)) {
                    thisobj = WScreenId(WScreenId::PlanetScreen, thisid);
                    thisq = 30;
                } else {
                    thisobj = WScreenId(global_anyplanet_type.getObject(thisid).getPos());
                    thisq = 31;
                }
            }
            break;
         case MessageIonStorm:
            if (global_ion_type.isExistant(thisid)) {
                thisobj = WScreenId(global_ion_type.getObject(thisid).getPos());
                thisq = 10;
            }
            break;
         case MessageExplosion:
         case MessageConfig:
         case MessagePlayerScore:
         case MessageNoObject:
            break;
        }

        if (thisobj.isNonNull() && thisq > q) {
            setHaveObject(true, thisobj);
            q = thisq;
        }
    }
}

WInboxWindow::WInboxWindow(GInbox& inbox)
    : UIWindow(_("Messages"), BLUE_WINDOW, COLOR_BLACK)
{
    init(inbox);
}

void
WInboxWindow::init(GInbox& inbox)
{
    // HBox
    //   VBox
    //     WMessageActionPanel
    //     HBox
    //       UIButton "Close"
    //       UISpacer
    //   WMessageDisplay
    // FIXME: this dialog currently has no clickable "Help" button.
    setLayout(UIHBoxLayout::instance5);

    UIGroup& g1  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g11 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new WInboxActionPanel(inbox)));
    g1.add(g11);

    g11.add(h.add(new UIButton(0, _("Close"), SDLK_ESCAPE)).addStop());
    g11.add(h.add(new UIButton(_("Tab - List"), SDLK_TAB)));
    g11.add(h.add(new UISpacer()));

    add(g1);
    add(h.add(new WMessageDisplay(0, inbox)));
    add(h.add(new UIPrefixArg()));
    add(h.add(new WHelpWidget("pcc2:msgin")));

    pack();
}

void
doInboxScreen()
{
    GInbox& inbox = getDisplayedTurn().getInbox();

    WInboxWindow win(inbox);
    ui_root->centerWidget(win);
    ui_root->runModal(win);
}
