/**
  *  \file client/globalact.cc
  *  \brief Global Actions Core
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "client/globalact.h"
#include "game/error.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "int/interror.h"

/** Constructor.
    \param search_result Reference to search result to use. Its lifetime must exceed
           the WGlobalActionParameter's. */
WGlobalActionParameter::WGlobalActionParameter(GObjectList& search_result)
    : flags(ExceptNumericalFCodes | ExceptSpecialFCodes | OnlyShips | OnlyPlanets),
      search_result(search_result)
{
    if (search_result.getNumObjects() > 0) {
        flags |= OnlySearchResult;
    }
}

WGlobalActionParameter::~WGlobalActionParameter()
{ }

/** Check whether an option (or at least one of a set of options) is set. */
bool
WGlobalActionParameter::hasFlag(int flag) const
{
    return (flags & flag) != 0;
}

/** Toggle an option. */
void
WGlobalActionParameter::toggleFlag(int flag)
{
    flags ^= flag;
    sig_changed.raise();
}

/** Set flag to a given value. Note that this only works if flag actually is
    a single flag. */
void
WGlobalActionParameter::setFlag(int flag, bool value)
{
    if (value != hasFlag(flag)) {
        toggleFlag(flag);
    }
}

/** Get search result. */
GObjectList&
WGlobalActionParameter::getSearchResult()
{
    return search_result;
}

/** Execute a global action.
    \param action Action to execute */
void
WGlobalActionParameter::execute(WGlobalAction& action)
{
    /* Prepare */
    std::auto_ptr<WGlobalActionExecuter> exec(action.prepare());
    if (!exec.get()) {
        return;
    }

    /* Execute */
    int objs = 0;
    int success = 0;

    // clock_t a = clock();

    if (hasFlag(OnlySearchResult)) {
        /* Iterate through search result */
        for (GObjectList::iterator i = search_result.begin(); i != search_result.end(); ++i) {
            if (i->isGameObject()) {
                maybeProcess(objs, success, *exec, i->getObject());
            }
        }
    } else {
        /* Iterate through whole universe */
        for (int i = global_anyship_type.getNextIndex(0); i != 0; i = global_anyship_type.getNextIndex(i)) {
            maybeProcess(objs, success, *exec, global_anyship_type.getObject(i));
        }
        for (int i = global_anyplanet_type.getNextIndex(0); i != 0; i = global_anyplanet_type.getNextIndex(i)) {
            maybeProcess(objs, success, *exec, global_anyplanet_type.getObject(i));
        }
    }

    // clock_t b = clock();

    /* Finish */
    getDisplayedTurn().getCurrentUniverse().doScreenUpdates();

    // clock_t c = clock();
    // printf("process %d objs: %f, update %f\n", objs, double(b-a)/CLOCKS_PER_SEC, double(c-b)/CLOCKS_PER_SEC);

    exec->finish(objs, success);
}

/** Process an object.
    \param objs [in/out] Found objects counter
    \param success [in/out] Processed objects counter
    \param exec Action executer to use
    \param obj Object to work on */
void
WGlobalActionParameter::maybeProcess(int& objs, int& success, WGlobalActionExecuter& exec, GObject& obj)
{
    /* Check basic status */
    if (!obj.isExistant()) {
        return;
    }
    if (hasFlag(OnlyMarked) && !obj.isMarked()) {
        return;
    }

    /* Check type, obtain fcode */
    string_t fc = "???";
    if (GShip* sh = dynamic_cast<GShip*>(&obj)) {
        if (!hasFlag(OnlyShips)) {
            return;
        }
        if (sh->isFCodeKnown()) {
            fc = sh->getFCode();
        }
    } else if (GPlanet* pl = dynamic_cast<GPlanet*>(&obj)) {
        if (!hasFlag(OnlyPlanets)) {
            return;
        }
        if (pl->isFCodeKnown()) {
            fc = pl->getFCode();
        }
    } else {
        return;
    }

    /* Check fcode */
    if (hasFlag(ExceptNumericalFCodes) && GFCode::isNumeric(fc)) {
        return;
    }
    if (hasFlag(ExceptSpecialFCodes) && (GFCode::isSpecial(fc, false) || GFCode::isExtraFC(fc))) {
        return;
    }

    /* OK, accept. Execute the operation. Swallow game errors. */
    ++objs;
    try {
        if (exec.execute(obj)) {
            ++success;
        }
    }
    catch (IntError&) { }
    catch (GError&) { }
}
