/**
  *  \file client/playerscr.cc
  *  \brief Player Main Screen
  *
  *  Main screen, user interface and multiplexing to other control screens.
  *  PCC 1.x source: cc.pas, scrmux.pas.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/playerscr.h"
#include "client/chartusr.h"
#include "client/dialogs/commandedit.h"
#include "client/dialogs/consoledlg.h"
#include "client/dialogs/teamedit.h"
#include "client/scr-score.h"
#include "client/scr-sim.h"
#include "client/screens.h"
#include "client/vcrplay.h"
#include "client/widgets/decaymsg.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/racename.h"
#include "game/ship.h"
#include "game/sim-game.h"
#include "game/stat.h"
#include "game/team.h"
#include "gfx/font.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "int/if/userif.h"
#include "int/interror.h"
#include "int/process.h"
#include "int/simple.h"
#include "resmgr/resmgr.h"
#include "ui/layout.h"
#include "ui/ui.h"
#include "ui/util.h"
#include "util/nls.h"
#include "ui/widgets/button.h"
#include "client/dialogs/imperial.h"
#include "client/dialogs/helpdlg.h"
#include "client/widgets/helpwidget.h"
#include "client/dialogs/gadialog.h"
#include "client/dialogs/histselect.h"
#include "client/screens/histscreen.h"

static void doScreen(WScreenId state);

class WPixmapButton : public UIButton {
    Ptr<GfxPixmap> pix;
 public:
    WPixmapButton(int aid, string_t aname, UIKey akey, Ptr<GfxPixmap> apix);
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
};

WPixmapButton::WPixmapButton(int aid, string_t aname,
                             UIKey akey, Ptr<GfxPixmap> apix)
    : UIButton(aid, aname, akey), pix(apix)
{ }

void
WPixmapButton::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_SMALL]);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    int text_height = ctx.getFont().getTextHeight(getName());
    if (pix && extent.h > text_height-2) {
        pix->blit(can, extent.x + 1, extent.y + 1,
                  GfxRect(0, 0, extent.w - 2, extent.h - text_height - 2));
    }
    ctx.setSolidBackground();   /* background is color 0 i.e. black */
    ctx.setColor(COLOR_WHITE);
    outTextF(ctx,
             extent.x + 1, extent.y + extent.h - text_height - 1,
             extent.w - 2, getName());
    drawFrameDown(can, extent);
    can.update(extent);
}

void
WPixmapButton::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint pt (fonts[FONT_SMALL]->getTextWidth(getName()) + 2,
                 fonts[FONT_SMALL]->getTextHeight(getName()) + 2);
    if (pix) {
        pt.x = pix->getWidth() + 2;
        pt.y += pix->getHeight();
    }
    info.flags = 0;
    info.min_size = info.pref_size = pt;
}

/***************************** WPlayerScreen *****************************/

class WPlayerScreen : public UIGroup, public IntUserInterfaceBinding {
    Ptr<GfxPixmap> pix;
    ObjectHolder holder;
    UIHBoxLayout layout;
    UIVBoxLayout vlayout;

    enum {
        kNone,
        kScreen,
        kConsole
    } script_action;
    WScreenId script_screen;
    IntKeymapRef keymap;

 public:
    WPlayerScreen();
    void drawContent(GfxCanvas& can);
    bool handleEvent(const UIEvent& e, bool second_pass);

    // IntUserInterfaceBinding:
    int  getProperty(IntUserInterfaceProperty p);
    void doGoToScreen(int screen, int obj);
    void doGoToChart(GPoint pt);
    void doMoveScanner(GPoint pt);
    void doPopupConsole();
    void doOverlayMessage(string_t msg);
    void doLeaveRace();
    void doSetChartView(string_t name);
    IntKeymapRef getKeymap();

    void doScripts(bool fromConsole = false);
};

WPlayerScreen::WPlayerScreen()
    : UIGroup(UINullLayout::instance /* will be overridden */), layout(5, 5), vlayout(5, 5),
      script_action(kNone),
      keymap(IntKeymap::getKeymapByName("RACESCREEN"))
{
    setLayout(layout);
    setExtent(GfxRect(*ui_screen));
    pix = requestPixmap("bgtile");
    UIGroup& g = holder.add(new UIGroup(vlayout));
    g.add(holder.add(new WPixmapButton(1, _("F1 - Starships"), SDLK_F1,
                             requestPixmap("menu.ship"))).addKey());
    g.add(holder.add(new WPixmapButton(2, _("F2 - Planets"), SDLK_F2,
                             requestPixmap("menu.planet"))).addKey());
    g.add(holder.add(new WPixmapButton(3, _("F3 - Starbases"), SDLK_F3,
                             requestPixmap("menu.base"))).addKey());
    g.add(holder.add(new WPixmapButton(4, _("F4 - Starchart"), SDLK_F4,
                             requestPixmap("menu.chart"))).addKey());
    g.add(holder.add(new UISpacer()));
    add(g);
    g.pack();

    UIGroup& fgroup = holder.add(new UIGroup(holder.add(new UIFlowLayout(7, true))));
    fgroup.add(holder.add(new UIButton(_("F7 - Search"), SDLK_F7)));
    fgroup.add(holder.add(new UIButton(_("G - Global"), 'g')));
    fgroup.add(holder.add(new UIButton(_("T - Teams"), 't')));
    fgroup.add(holder.add(new UIButton(_("S - Scores"), 's')));
    fgroup.add(holder.add(new UIButton(_("B - Battle Simulator"), 'b')));
    fgroup.add(holder.add(new UIButton(_("I - Imperial Stats"), 'i')));
    fgroup.add(holder.add(new UIButton(_("W - Write Message"), 'w')));
    fgroup.add(holder.add(new UIButton(_("M - Messages"), 'm')).setState(st_Disabled, getDisplayedTurn().getInbox().getCount() == 0));
    fgroup.add(holder.add(new UIButton(_("V - Combat Recorder"), 'v')).setState(st_Disabled, getDisplayedTurn().getVcrs().getCount() == 0));
    fgroup.add(holder.add(new UIButton(_("ESC - Exit"), SDLK_ESCAPE)));
    fgroup.add(holder.add(new UIButton(_("H - Help"), 'h')));

    UIGroup& vgroup = holder.add(new UIGroup(vlayout));
    vgroup.add(holder.add(new UISpacer()));
    vgroup.add(fgroup);
    add(vgroup);

    doLayout();
}

void
WPlayerScreen::drawContent(GfxCanvas& can)
{
    drawTiledArea(can, GfxRect(extent.x + 1, extent.y + 1,
                               extent.w - 2, extent.h - 2),
                  pix, standard_colors[COLOR_DARK], 16);
    drawFrameUp(can, extent);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setColor(COLOR_WHITE);

    outText(ctx, 200,  20, player_racenames.getFullName(getPlayerId()));
    outText(ctx, 200,  60, format(_("Turn %d"), getDisplayedTurn().getTurnNumber()));
    outText(ctx, 200,  80, format(_("%d message%!1{s%}"), getDisplayedTurn().getInbox().getCount()));
    outText(ctx, 200, 100, format(_("%d combat recording%!1{s%}"), getDisplayedTurn().getVcrs().getCount()));

    can.update(extent);
}

/** Play VCRs. */
static void
playVcrs(GVcrDatabase& db)
{
    // HACK HACK HACK
    GClassicVcrDatabase* cdb = dynamic_cast<GClassicVcrDatabase*>(&db);
    if (!cdb)
        return;

    ObjectHolder holder;
    WVcrSelector sel(*cdb);
    WVcrSelectorWindow w(sel);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(holder.add(new UIButton(cm_VCR_Prev, "\xE2\x86\x91", '-')).addCommand());
    but.add(holder.add(new UIButton(cm_VCR_Next, "\xE2\x86\x93", '+')).addCommand());
    but.add(holder.add(new UIButton(cm_Enter, _("ENTER - Play"), SDLK_RETURN)).addCommand());
    but.add(holder.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE)).addStop());
    but.add(holder.add(new UISpacer()));
    but.add(holder.add(new UIButton(_("Help"), 'h')));

    w.setLayout(UIVBoxLayout::instance5);
    w.add(sel);
    w.add(but);
    w.add(holder.add(new WHelpWidget("pcc2:vcr")));
    w.pack();
    ui_root->centerWidget(w);
    ui_root->add(w);
    ui_root->run();
    ui_root->close(w);
}

int
WPlayerScreen::getProperty(IntUserInterfaceProperty p)
{
    switch (p) {
     case iuiScreenNumber:
     case iuiSimFlag:
     case iuiScreenRegistered:
        return 0;
     case iuiChartX:
     case iuiScanX:
        return getStarchartPosition().x;
     case iuiChartY:
     case iuiScanY:
        return getStarchartPosition().y;
    }
    return -1;
}

void
WPlayerScreen::doGoToScreen(int screen, int obj)
{
    if (screen == 0) {
        // Fine, we're already there
    } else if (screen == 4) {
        script_action = kScreen;
        script_screen = WScreenId(getStarchartPosition());
    } else {
        script_action = kScreen;
        script_screen = WScreenId(WScreenId::ScreenId(screen), obj);
    }
}

void
WPlayerScreen::doGoToChart(GPoint pt)
{
    script_action = kScreen;
    script_screen = WScreenId(pt);
}

void
WPlayerScreen::doMoveScanner(GPoint pt)
{
    setStarchartPosition(pt);
}

void
WPlayerScreen::doPopupConsole()
{
    script_action = kConsole;
}

void
WPlayerScreen::doOverlayMessage(string_t msg)
{
    showDecayingMessage(msg);
}

void
WPlayerScreen::doLeaveRace()
{
    stop(-1);
}

void
WPlayerScreen::doSetChartView(string_t /*name*/)
{
    throw IntError::contextError();
}

IntKeymapRef
WPlayerScreen::getKeymap()
{
    return keymap;
}

void
WPlayerScreen::doScripts(bool fromConsole)
{
    do {
        fromConsole = false;
        switch (script_action) {
         case kScreen:
            script_action = kNone;
            doScreen(script_screen);
            fromConsole = true;
            break;
         case kConsole:
            script_action = kNone;
            doConsole();
            fromConsole = true;
            break;
         case kNone:
            break;
        }
        if (script_action == kNone) {
            runRunnableProcesses();
            killTerminatedProcesses();
        }
    } while (script_action != kNone);
}

bool
WPlayerScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        if (keymap != 0) {
            uint32_t cmd = keymap->lookupCommand(e.key.code);
            if (isAtom(cmd)) {
                executeStatement(atomStr(cmd), e.key.arg);
                doScripts(true /* means: do not run scripts, first check results */);
                return true;
            }
        }
        if (e.key.code == 'v') {
            if (getDisplayedTurn().getVcrs().getCount())
                playVcrs(getDisplayedTurn().getVcrs());
            return true;
        } else if (e.key.code == 'b') {
            // FIXME: this should be converted to use WScreenId as well.
            // This is not trivial as GSimulatorRealGameInterface is a 'game'
            // class and does not know about the WScreenId 'client' class.
            // A possible method would be to implement an own descendant of
            // GSimulatorRealGameInterface which overrides selectGameShip().
            GSimulatorRealGameInterface rgi(getDisplayedTurn().getCurrentUniverse());
            doScreen(WScreenId(WScreenId::ScreenId(doSimulator(rgi)), 0));
            doScripts();
            return true;
        } else if (e.key.code == 's') {
            if (isStatisticsFileLoaded() && getStatistics().getNumTurns() != 0)
                doScoreScreen(getStatistics(), getPlayerId());
            return true;
        } else if (e.key.code == 't') {
            editTeams();
            return true;
        } else if (e.key.code == 'i') {
            doImperialStatistics();
            doScripts();
            return true;
        } else if (e.key.code == 'g') {
            doGlobalActions();
            doScripts();
            return true;
        } else if (e.key.code == ss_Alt + 'e') {
            editCommands();
            doScripts();
            return true;
        }
    }
    return UIGroup::handleEvent(e, second_pass);
}

/***************************** Main Routines *****************************/

void
doPlayerScreen()
{
    WPlayerScreen ps;
    ui_root->add(ps);
    ps.doScripts();
    ui_root->run();
    ui_root->close(ps);
}

/** Display a screen. This is no longer a loop as it used to be,
    looping is done impliclty; the control screens post script events
    containing the respective next screen. */
static void
doScreen(WScreenId state)
{
    switch (state.getScreenId()) {
     case WScreenId::ShipScreen:
        if (state.getId())
            global_ship_type.setCurrent(state.getId());
        doShipScreen();
        break;

     case WScreenId::PlanetScreen:
        if (state.getId())
            global_planet_type.setCurrent(state.getId());
        doPlanetScreen();
        break;

     case WScreenId::BaseScreen:
        if (state.getId())
            global_base_type.setCurrent(state.getId());
        doBaseScreen();
        break;

     case WScreenId::HistoryScreen:
        if (state.getId())
            global_histship_type.setCurrent(state.getId());
        doHistoryScreen();
        break;

     case WScreenId::Starchart:
        setStarchartPosition(state.getPos());
        doStarchart();
        break;

     default:
        break;
    }
}
