/**
  *  \file client/widgets/histlist.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include "client/widgets/histlist.h"
#include "game/team.h"
#include "game/spec.h"
#include "game/ship.h"
#include "game/univ.h"
#include "game/objlist.h"
#include "game/global.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "gfx/gfxcontext.h"
#include "ui/skin.h"
#include "game/format.h"

/** Build list of history ships. */
void
WHistoryShipSelection::buildList(GObjectList& list, const GUniverse& univ) const
{
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        const GShip& sh = univ.getShip(sid);
        bool accept = false;
        if (sh.getKind() != GShip::NoShip) {
            switch (mode) {
             case AllShips:
                accept = true;
                break;
             case LocalShips:
                for (int t = sh.getHistoryNewestTurn(); t > sh.getHistoryOldestTurn(); --t) {
                    const TDbShipTrackEntry e = sh.getHistoryEntry(t);
                    if (e.x > 0 && e.y > 0 && std::abs(e.x - location.x) < 10 && std::abs(e.y - location.y) < 10) {
                        accept = true;
                        break;
                    }
                }
                break;
             case ForeignShips:
                accept = (sh.getOwner() != getPlayerId());
                break;
             case TeamShips:
                accept = (getPlayerTeam(sh.getOwner()) == getPlayerTeam(getPlayerId()));
                break;
             case EnemyShips:
                accept = (getPlayerTeam(sh.getOwner()) != getPlayerTeam(getPlayerId()));
                break;
             case OwnShips:
                accept = (sh.getOwner() == getPlayerId());
                break;
            }
        }
        if (accept) {
            list.addObject(global_histship_type, sid);
        }
    }
}

/** Get number of last turn a ship was seen.
    \param univ Universe to query
    \param sid Ship Id. Must be among the ships returned by buildList.
    \return turn number, 0 if none */
int
WHistoryShipSelection::getShipLastTurn(const GUniverse& univ, int sid) const
{
    const GShip& sh = univ.getShip(sid);
    for (int t = sh.getHistoryNewestTurn(); t > sh.getHistoryOldestTurn(); --t) {
        const TDbShipTrackEntry e = sh.getHistoryEntry(t);
        if (e.x > 0 && e.y > 0
            && (mode != LocalShips
                || (std::abs(e.x - location.x) < 10 && std::abs(e.y - location.y) < 10)))
        {
            return t;
        }
    }
    return 0;
}

/** Get set of visible modes. */
SmallSet<WHistoryShipSelection::Mode>
WHistoryShipSelection::getVisibleModes(const GUniverse& univ) const
{
    SmallSet<Mode> modes;
    SmallSet<Mode> expect = SmallSet<Mode>::allUpTo(Mode(ModeMax));
    bool expectLocalShips = (location.x != 0 || location.y != 0);
    if (!expectLocalShips) {
        expect -= LocalShips;
    }

    /* Check all ships */
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        const GShip& sh = univ.getShip(sid);
        if (sh.getKind() != GShip::NoShip) {
            /* We have a ship. Process owner modes */
            modes += AllShips;
            if (sh.getOwner() == getPlayerId()) {
                modes += OwnShips;
            } else {
                modes += ForeignShips;
            }
            if (getPlayerTeam(sh.getOwner()) == getPlayerTeam(getPlayerId())) {
                modes += TeamShips;
            } else {
                modes += EnemyShips;
            }

            /* Check for local mode */
            if (expectLocalShips) {
                for (int t = sh.getHistoryNewestTurn(); t > sh.getHistoryOldestTurn(); --t) {
                    const TDbShipTrackEntry e = sh.getHistoryEntry(t);
                    if (e.x > 0 && e.y > 0 && std::abs(e.x - location.x) < 10 && std::abs(e.y - location.y) < 10) {
                        modes += LocalShips;
                        break;
                    }
                }
            }

            /* Check for early exit when full set found */
            if (modes == expect) {
                break;
            }
        }
    }

    /* Discount team modes when we don't have teams */
    if (!isTeamConfigured()) {
        modes -= TeamShips;
        modes -= EnemyShips;
    }
    return modes;
}

WHistoryShipSelection::Mode
WHistoryShipSelection::getInitialMode(const GUniverse& univ)
{
    SmallSet<Mode> modes = getVisibleModes(univ);
    if (modes.contains(LocalShips)) {
        return LocalShips;
    } else {
        return AllShips;
    }
}


string_t
WHistoryShipSelection::getModeName(Mode mode)
{
    switch (mode) {
     case AllShips:
        return _("All ships");
     case LocalShips:
        return format(_("Ships near (%d,%d)"), location.x, location.y);
     case ForeignShips:
        return _("Foreign ships");
     case TeamShips:
        return _("Team ships");
     case EnemyShips:
        return _("Enemy ships");
     case OwnShips:
        return _("Own ships");
    }
    return string_t();
}

string_t
WHistoryShipSelection::getModeName()
{
    return getModeName(mode);
}

/**************************** WHistoryShipList ***************************/

/** Constructor.
    \param id     Widget Id
    \param lines  Lines to display
    \param width  Width of widget
    \param list   (Storage for) ship list
    \param config Initial configuration */
WHistoryShipList::WHistoryShipList(int id, int lines, int width, GObjectList& list, WHistoryShipSelection& config)
    : WObjectSelectionList(id, lines, width, list),
      config(config),
      list(list)
{
    setInfoSpace(fonts[FONT_NORMAL]->getTextWidth(format(_("%d turns ago"), 100)));
}

WHistoryShipList::~WHistoryShipList()
{ }

string_t
WHistoryShipList::getObjectInfo(GObjectId id)
{
    // Additional information, used by WObjectSelectionList::drawPart
    GShip* sh;
    if (id.isValid() && (sh = dynamic_cast<GShip*>(&id.getObject())) != 0) {
        int turn = config.getShipLastTurn(getDisplayedTurn().getCurrentUniverse(), sh->getId());
        if (turn > 0) {
            return formatTurnNumber(turn);
        }
    }
    return string_t();
}

/** Update list. Call this after a change to the configuration to
    update the display. This is intended to be hooked into
    WHistoryShipControl::sig_changed. */
void
WHistoryShipList::updateList()
{
    /* Remember current object */
    GObjectId current = getCurrentObject();

    /* Build new list */
    list.clear();
    config.buildList(list, getDisplayedTurn().getCurrentUniverse());

    /* Redraw */
    onListChange(current);
}


/************************** WHistoryShipControl **************************/

WHistoryShipControl::WHistoryShipControl(WHistoryShipSelection& config)
    : UIAbstractButton(0, ss_Alt + 'd'),
      config(config)
{
    sig_fired.add(this, &WHistoryShipControl::onClick);
}

WHistoryShipControl::~WHistoryShipControl()
{ }

void
WHistoryShipControl::onClick()
{
    /* Toggle mode */
    SmallSet<WHistoryShipSelection::Mode> modes = config.getVisibleModes(getDisplayedTurn().getCurrentUniverse());
    WHistoryShipSelection::Mode mode = config.mode;
    for (int i = 0; i < WHistoryShipSelection::ModeMax; ++i) {
        /* Find next mode */
        if (mode == WHistoryShipSelection::Mode(WHistoryShipSelection::ModeMax)) {
            mode = WHistoryShipSelection::Mode(0);
        } else {
            mode = WHistoryShipSelection::Mode(mode + 1);
        }
        /* Found a valid mode? Update everything. */
        if (modes.contains(mode)) {
            config.mode = mode;
            sig_changed.raise();
            drawWidget();
            break;
        }
    }
}

void
WHistoryShipControl::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);
    ctx.setSolidBackground();
    ctx.setTextAlign(0, 1);

    const GfxRect& ex = getExtent();
    outTextF(ctx, ex.x, ex.y + ex.h/2, ex.w, config.getModeName());

    can.update(ex);
}

void
WHistoryShipControl::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = font_sizes[FONT_NORMAL].scaledBy(15, 1);
    info.flags = li_FlexibleH;
}
