/**
  *  \file game/stat.h
  *  \brief Score Statistics - Management and I/O
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_STAT_H
#define PCC_V2_GAME_STAT_H

#include <vector>
#include "cpluslib/types.h"
#include "game/struct.h"
#include "game/timestamp.h"
#include "cpluslib/ptrvector.h"

class Stream;
class GGen;
class GMessageInformation;

typedef unsigned int GStatIndex;

enum {
    ScoreId_Planets      = -1,
    ScoreId_Capital      = -2,
    ScoreId_Freighters   = -3,
    ScoreId_Bases        = -4,

    ScoreId_Score        = 1,
    ScoreId_BuildPoints  = 2,
    ScoreId_MinesAllowed = 3,
    ScoreId_MinesLaid    = 4
};

/** Score file record. Contains one turn's scores, as a two-dimensional array (see operator()).
    The indexes are variable according to the GStatFile's schema. */
class GStatRecord {
    int                turn;
    GTimestamp         timestamp;
    std::vector<int32> scores;

 public:
    GStatRecord(int turn, GTimestamp time, uint16 num_fields);

    ~GStatRecord();

    bool loadData(Stream& s);

    bool save(Stream& s) const;

    int getTurnNumber() const;

    GTimestamp const& getTimestamp() const;

    int32& operator()(GStatIndex i, int player);

    int32 operator()(GStatIndex i, int player) const;

    void setNumberOfFields(uint16 new_count);
};

/** Statistics file reading. PCC 1.x does lazy parsing of statistics files,
    and I originally planned that for PCC2 as well. The file format is
    designed to support such a type of access. However, this would have meant
    quite some complexity and several states to distinguish between all
    operations that can possibly be done on a statistics file. Since
    the file will be a few k only, it makes sense to simply slurp it into
    memory completely and write it out completely as well. To avoid losing
    files written by a future version, we manage a flag saying so. */
class GStatFile {
    TScoreHeader header;

    std::vector<int16> record_defs;

    std::vector<TScoreDescription> score_descriptions;

    bool file_used_future_features;

    ptr_vector<GStatRecord> loaded_records;

    void clear();
 public:

    static const GStatIndex nil = GStatIndex(-1);

    GStatFile();

    ~GStatFile();

    void load(Stream& s);

    void loadOldFile(Stream& s);

    void init();

    void save(Stream& s);

    bool canSaveSafely() const;

    // Access to schema definition and descriptions:
    GStatIndex getNumScores() const;

    GStatIndex getIndexForScore(int16 score_id) const;

    GStatIndex getAddIndexForScore(int16 score_id);

    int16 getScoreId(GStatIndex index) const;

    const TScoreDescription* getDescriptionForScoreId(int16 score_id) const;

    bool addScoreDescription(const TScoreDescription& desc);

    // Access to data:
    uint32 getNumTurns() const;

    GStatRecord& getRecordForTurn(int turn, const GTimestamp& ts);

    GStatRecord& getRecordByIndex(uint32 index);
    const GStatRecord& getRecordByIndex(uint32 index) const;

    // Data merging:
    void mergeGen(const GGen& g);
    void addMessageInformation(const GMessageInformation& mi, const GTimestamp& ts);
};


bool isStatisticsFileLoaded();
GStatFile& getStatistics();

extern const struct ModuleInitializer mod_stat[];


/** Get turn number. */
inline int
GStatRecord::getTurnNumber() const
{
    return turn;
}

/** Get timestamp of this turn. */
inline GTimestamp const&
GStatRecord::getTimestamp() const
{
    return timestamp;
}

/** Access to an entry.
    \param i Index obtained from GStatFile::getIndexForScore()
    \param player player number, [1,NUM_PLAYERS] */
inline int32&
GStatRecord::operator()(GStatIndex i, int player)
{
    return scores[NUM_PLAYERS*i + player-1];
}

/** Access to an entry.
    \param i Index obtained from GStatFile::getIndexForScore()
    \param player player number, [1,NUM_PLAYERS] */
inline int32
GStatRecord::operator()(GStatIndex i, int player) const
{
    return scores[NUM_PLAYERS*i + player-1];
}



/** Check whether we can safely save this file without losing information.
    If the file we loaded contained some information we don't understand, refuse
    to save the file to avoid data loss. */
inline bool
GStatFile::canSaveSafely() const
{
    return !file_used_future_features;
}

/** Get number of scores. This is the number of "columns" the table has; usually
    five (planets/capital/freighters/bases/PBPs) but may be more. */
inline GStatIndex
GStatFile::getNumScores() const
{
    return record_defs.size();
}

/** Get score for an index.
    \param index Index, [0,getNumScores())
    \return score type, ScoreId_Xxx or different */
inline int16
GStatFile::getScoreId(GStatIndex index) const
{
    return record_defs[index];
}

/** Get number of turns stored. */
inline uint32
GStatFile::getNumTurns() const
{
    return loaded_records.size();
}

/** Get turn record by index.
    \param index Index, [0,getNumTurns()). */
inline GStatRecord&
GStatFile::getRecordByIndex(uint32 index)
{
    return *loaded_records[index];
}

/** Get turn record by index.
    \param index Index, [0,getNumTurns()). */
inline const GStatRecord&
GStatFile::getRecordByIndex(uint32 index) const
{
    return *loaded_records[index];
}

#endif
