/**
  *  \file io/dirs.cc
  *  \brief Directory Handling
  *
  *  Directory reading and pathname arithmetic. Most of the functions
  *  are platform-dependant and implemented in arch/.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#include "arch/platform.h"
#include "cpluslib/assert.h"
#include "cpluslib/smartptr.h"
#include "io/dirs.h"
#include "util/except.h"
#include "util/nls.h"

/**
  *  \class DirectoryEntry
  *  \brief Abstract Directory Entry
  *
  *  This class provides an interface for directory entries.
  *  It can be opened to obtain the actual file stream or
  *  subdirectory.
  */

/** Constructor. Construct a directory entry for the specified
    name. The name should be the complete name of the file. */
DirectoryEntry::DirectoryEntry(string_t name)
    : name(name), info_set(false)
{ }

DirectoryEntry::~DirectoryEntry()
{ }

/** Get size of file. Returns the size of the file in bytes,
    or -1 when an error occurs. Results are cached, that is,
    when you query the directory item the first time, the
    disk is actually read; later on the cached result is used. */
DirectoryEntry::fpos_t
DirectoryEntry::getSize()
{
    if (!info_set)
        getInfo();
    if (info_set)
        return file_size;
    else
        return -1;
}

/** Get file type. Returns the type of the file:
    - FT_FILE for regular files
    - FT_DIRECTORY for subdirectories
    - FT_UNKNOWN when the type can't be determined
    or some other implementation defined value. Results are cached
    like for getSize(). */
DirectoryEntry::FileType
DirectoryEntry::getFileType()
{
    if (!info_set)
        getInfo();
    if (info_set)
        return file_type;
    else
        return FT_UNKNOWN;
}

/** Set type/size information. Your implementation of getInfo()
    should call this method, filling in the appropriate values
    for size and type. */
void
DirectoryEntry::setInfo(fpos_t size, FileType type)
{
    file_size = size;
    file_type = type;
    info_set  = true;
}

/** Forget stored information. */
void
DirectoryEntry::clearInfo()
{
    file_size = 0;
    file_type = 0;
    info_set = false;
}

/** Get basename. Should return the file name with all path information
    removed. \default does standard pathname arithmetic. You should only
    override it if you can get the same information much more efficient. */
string_t
DirectoryEntry::getBaseName() const
{
    return Directory::getFileName(getName());
}

/************************** DummyDirectoryEntry **************************/

/**
  *  \class DummyDirectoryEntry
  *  \brief Dummy Entry
  *
  *  This is a directory entry usable for `dead ends'. It
  *  can not be opened in any way.
  */

/** Constructor.
    \param name Name of item
    \param parent Parent directory. Must not be 0. */
DummyDirectoryEntry::DummyDirectoryEntry(string_t name, Ptr<Directory> parent)
    : DirectoryEntry(name), parent(parent)
{
    ASSERT(parent.ptr());
}

Ptr<Directory>
DummyDirectoryEntry::openDir()
{
    throw FileOpenException(getName(), _("No such directory"));
}

Ptr<Stream>
DummyDirectoryEntry::openFile(int)
{
    throw FileOpenException(getName(), _("No such file"));
}

Ptr<Directory>
DummyDirectoryEntry::openParent()
{
    return parent;
}

void
DummyDirectoryEntry::erase()
{
    throw FileOpenException(getName(), _("No such file"));
}

void
DummyDirectoryEntry::renameTo(string_t)
{
    throw FileOpenException(getName(), _("No such file"));
}

void
DummyDirectoryEntry::createAsDirectory()
{
    throw FileOpenException(getName(), _("Cannot create directory"));
}

void
DummyDirectoryEntry::getInfo()
{ }

/******************************* Directory *******************************/

/**
  *  \class Directory
  *  \brief Abstract Directory
  *
  *  This class provides an abstract interface for directories.
  *  Directories can be queried for directory items (i.e. the
  *  files they contain); these are presented as instances of
  *  classes derived from DirectoryEntry.
  *
  *  In addition, there are functions for OS-dependant file name
  *  `arithmetic'.
  */

Directory::~Directory()
{ }

/** Open subdirectory. This is just a convenience function for
    DirectoryEntry::openDir(). */
Ptr<Directory>
Directory::openSubdir(string_t name)
{
    return getEntryByName(name)->openDir();
}

/** Open file. This is just a convenience function for
    DirectoryEntry::openFile(int). */
Ptr<Stream>
Directory::openFile(string_t name, int mode)
{
    return getEntryByName(name)->openFile(mode);
}

/** Open file, don't throw. When the file can't be opened, returns 0
    instead of throwing. \see openFile */
Ptr<Stream>
Directory::openFileNT(string_t name, int mode)
{
    try {
        return getEntryByName(name)->openFile(mode);
    }
    catch(...) {
        return 0;
    }
}

/** Erase file. This is a convenience function for
    DirectoryEntry::erase(). */
void
Directory::erase(string_t name)
{
    return getEntryByName(name)->erase();
}

/** Erase file, don't throw. When the file can't be erased, returns
    false. On success, returns true. */
bool
Directory::eraseNT(string_t name)
{
    try {
        getEntryByName(name)->erase();
        return true;
    }
    catch(...) {
        return false;
    }
}

/** Get title of this directory. This is the name which should be
    presented to the user. Derived classes can override it if they want.
    \default returns basename. */
string_t
Directory::getTitle()
{
    return Directory::getFileName(getName());
}

/** Append extension to a filename.
    \param name  Name, possibly including directory
    \param ext   Extension, not including dot
    \param force Always replace extension, even if input already has one */
string_t
appendFileNameExtension(string_t name, string_t ext, bool force)
{
    string_t filename = Directory::getFileName(name);
    if (filename.size() == 0) {
        /* ugly pathological case */
        return Directory::makeFileName(Directory::getDirName(name), "." + ext);
    } else {
        /* do not accept index 0 to avoid identifying ".emacs" as
           zero-length basename with extension EMACS */
        string_t::size_type n = filename.rfind('.');
        if (n == filename.npos || n == 0) {
            return Directory::makeFileName(Directory::getDirName(name), filename + "." + ext);
        } else if (force) {
            return Directory::makeFileName(Directory::getDirName(name), filename.substr(0, n+1) + ext);
        } else {
            return name;
        }
    }
}
