/**
  *  \file reshack/fontedit.cc
  *  \brief Resource Editor: Font Editor
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "reshack/fontedit.h"
#include "cpluslib/format.h"
#include "gfx/palette.h"
#include "io/dirs.h"
#include "reshack/charname.h"
#include "reshack/colorsel.h"
#include "reshack/painter.h"
#include "reshack/select.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/except.h"
#include "util/nls.h"
#include "ui/msgbox.h"
#include "io/io.h"
#include "cpluslib/utf8.h"
#include "util/recode.h"
#include "ui/util.h"
#include "reshack/grid.h"
#include "reshack/fontdlg.h"
#include "reshack/fontsynth.h"
#include "ui/widgets/keytranslator.h"
#include "ui/widgets/inputline.h"
#include "ui/widgets/button.h"
#include "ui/widgets/radiobutton.h"

namespace {
    /** Font editor window. This is the main font editor screen,
        providing a painter and tool, color and character selection. */
    class RHFontEditorWindow : public UIWindow {
        ObjectHolder h;
        Ptr<GfxBitmapFont> font;
        uint32_t current_character;

        RHCharacterName* charname_widget;
        RHPainter* painter_widget;

        void init();
        void setCharacter(uint32_t ch);
     public:
        RHFontEditorWindow(Ptr<GfxBitmapFont> font, string_t title);
        ~RHFontEditorWindow();
        bool handleEvent(const UIEvent& event, bool second_pass);
        bool pickCharacter(string_t title, uint32_t& result, bool withGrid);
        void storeCurrentCharacter();
        Ptr<GfxBitmapFont> getFont()
            { return font; }
    };

    /** Character selection dialog. Provides a filterable list of
        character names. */
    class RHPickCharacterDialog : public UIWindow {
        ObjectHolder h;
        CharacterList list;
        RHCharacterNameList* list_widget;
        UIInputLine* input_widget;
        UIScrollbar* sb_widget;

        void init();
        void onType();
     public:
        RHPickCharacterDialog(string_t title);

        static bool doDialog(string_t title, uint32_t& result);
    };

    /** Character grid dialog. */
    class RHCharacterGridDialog : public UIWindow {
        ObjectHolder h;
        RHCharacterGrid* grid_widget;
        RHCharacterName* charname_widget;

        void init(const GfxBitmapFont& font);
        void onMove();
     public:
        RHCharacterGridDialog(const GfxBitmapFont& font, string_t title);

        static bool doDialog(const GfxBitmapFont& font, string_t title, uint32_t& result);
    };

    /** Character in system font widget. */
    class RHSystemCharacterWidget : public UIWidget {
        const uint32_t ch;
     public:
        RHSystemCharacterWidget(uint32_t ch);

        void drawContent(GfxCanvas& can);
        bool handleEvent(const UIEvent& event, bool second_pass);
        void getLayoutInfo(LayoutInfo& info);
    };

    /** Character generator. */
    class RHEditorCharacterGenerator : public RHCharacterGenerator {
        RHPainter& painter;
        GfxBitmapFont& font;
     public:
        RHEditorCharacterGenerator(RHPainter& painter, GfxBitmapFont& font);
        bool check(uint32_t chars[], uint32_t nchars);
    };

    /** "Save" dialog for fonts. Asks for a file name and format
        parameters. */
    class RHSaveFontDialog : public UIWindow {
        ObjectHolder h;
        Observable<int> encoding;
        UIInputLine file_name;
        UIInputLine font_name;

        void init();
     public:
        RHSaveFontDialog();
        ~RHSaveFontDialog();

        void setFileName(string_t fn)
            { file_name.setText(fn); }
        string_t getFileName() const
            { return file_name.getText(); }
        void setFontName(string_t fn)
            { font_name.setText(fn); }
        string_t getFontName() const
            { return font_name.getText(); }
        void setEncoding(int enc)
            { encoding.set(enc); }
        int getEncoding() const
            { return encoding.get(); }
    };
}

/** Convert a character glyph to an editable pixmap.
    \param fnt Font
    \param ch  Character number
    \return New pixmap. Never null. */
static Ptr<GfxPixmap>
glyphToPixmap(const GfxBitmapFont& fnt, uint32_t ch)
{
    const GfxBitmapGlyph* g = fnt.getGlyph(ch);

    // Create pixmap
    Ptr<GfxPixmap> pix = new GfxPixmap(g ? g->getWidth() : 0, fnt.getHeight(), 8, 0, 0, 0, 0);
    SDL_SetPalette(pix->getSystemSurface(), SDL_LOGPAL|SDL_PHYSPAL, standard_palette, 0, NUM_STANDARD_COLORS);

    // Populate pixmap
    if (g)
        g->drawColored(*pix, 0, 0, FC_WHITE, FC_HALF);

    return pix;
}

/** Convert pixmap to character glyph.
    \param pix Pixmap
    \return newly-allocated glyph. Can be 0 if pixmap is empty (zero-size). */
static GfxBitmapGlyph*
pixmapToGlyph(Ptr<GfxPixmap> pix)
{
    if (pix->getWidth() == 0) {
        // Empty image: return null glyph
        return 0;
    } else {
        // Non-empty image: build a glyph
        std::auto_ptr<GfxBitmapGlyph> g(new GfxBitmapGlyph(pix->getWidth(), pix->getHeight()));
        for (int y = 0; y < pix->getHeight(); ++y) {
            for (int x = 0; x < pix->getWidth(); ++x) {
                switch (pix->getPixel(GfxPoint(x, y))) {
                 case FC_BLACK:
                    break;
                 case FC_HALF:
                    g->addAAHint(x, y);
                    break;
                 case FC_WHITE:
                 default:
                    g->set(x, y, true);
                }
            }
        }
        return g.release();
    }
}

/** Get next existing (non-null) character from font.
    \param fnt Font
    \param ch  Starting character
    \param delta Direction (+1 or -1)
    \return New character number */
static uint32_t
getNextExistingCharacter(const GfxBitmapFont& fnt, uint32_t ch, int32_t delta)
{
    do {
        ch += delta;
    } while (ch != 0 && ch != 0xFFFF && fnt.getGlyph(ch) == 0);
    return ch;
}

static int
showCharacterInSystemFonts(uint32_t ch)
{
    // VBox
    //   RHSystemCharacterWidget
    //   HBox
    //     UISpacer
    //     UIButton "Close"
    UIWindow w(_("System Fonts"), UIWindow::BLUE_WINDOW, COLOR_BLACK);
    RHSystemCharacterWidget scw(ch);
    UIButton btn_close(-1, _("Close"), SDLK_ESCAPE);
    UISpacer spacer;
    UIGroup group(UIHBoxLayout::instance5);
    UIKeyTranslator kt;
    kt.addStop(SDLK_RETURN, -1);
    kt.addStop(' ', -1);

    w.setLayout(UIVBoxLayout::instance5);
    w.add(scw);
    w.add(group);
    w.add(kt);
    group.add(spacer);
    group.add(btn_close);
    btn_close.addStop();
    w.pack();
    ui_root->centerWidget(w);
    return ui_root->runModal(w);
}

/** Change encoding from some codepage to Unicode. */
static void
changeToUnicode(GfxBitmapFont& font, CharacterSet cs)
{
    // Remember all the glyphs
    ptr_vector<GfxBitmapGlyph> codepageGlyphs;
    for (int i = 128; i < 256; ++i) {
        const GfxBitmapGlyph* fontGlyph = font.getGlyph(i);
        if (fontGlyph)
            codepageGlyphs.push_back(new GfxBitmapGlyph(*fontGlyph));
        else
            codepageGlyphs.push_back(0);
    }

    // Delete all font glyphs
    for (int i = 128; i < 256; ++i)
        font.addNewGlyph(i, 0);

    // Now copy them to their proper place
    for (int i = 128; i < 256; ++i) {
        uint16_t uni = cs.getUnicode(i);
        if (codepageGlyphs[i-128] != 0 && uni >= 128) {
            font.addNewGlyph(uni, codepageGlyphs.extract_elem(i-128));
        }
    }
}

static int
findTop(const GfxBitmapGlyph& g)
{
    for (int y = 0; y < g.getHeight(); ++y)
        for (int x = 0; x < g.getWidth(); ++x)
            if (g.get(x, y))
                return y;
    return -1;
}

static int
findBottom(const GfxBitmapGlyph& g)
{
    for (int y = g.getHeight(); y > 0; --y)
        for (int x = 0; x < g.getWidth(); ++x)
            if (g.get(x, y-1))
                return y;
    return -1;
}

/*************************** RHFontEditorWindow **************************/

/** Constructor.
    \param font Font to edit
    \param title Window title */
RHFontEditorWindow::RHFontEditorWindow(Ptr<GfxBitmapFont> font, string_t title)
    : UIWindow(title, BLUE_WINDOW, COLOR_BLACK),
      font(font),
      current_character(0),
      charname_widget(0),
      painter_widget(0)
{
    init();
}

RHFontEditorWindow::~RHFontEditorWindow()
{ }

void
RHFontEditorWindow::init()
{
    // HBox
    //   VBox
    //     RHToolSelector
    //     RHZoomSelector
    //     UISpacer
    //   VBox
    //     HBox
    //       UIButton <
    //       RHCharacterName
    //       UIButton "G"
    //       UIButton >
    //     RHPainter
    //   VBox
    //     RHFontColorSelector
    //     UISpacer
    //     UIButton "close"

    UIGroup& g1  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g2  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g21 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g3  = h.add(new UIGroup(UIVBoxLayout::instance5));

    Ptr<GfxPixmap> pixmap = glyphToPixmap(*font, current_character);
    RHPainter& painter = h.add(new RHPainter(pixmap, GrayscaleColor));
    RHToolSelector& sel = h.add(new RHToolSelector(painter));
    sel.addNewTool(SDLK_F1, new RHPencilTool());
    sel.addNewTool(SDLK_F2, new RHBrushTool());
    sel.addNewTool(SDLK_F3, new RHLineTool());
    sel.addNewTool(SDLK_F4, new RHRectangleTool());
    sel.addNewTool(SDLK_F5, new RHBarTool());
    sel.addNewTool(SDLK_F6, new RHCircleTool());
    sel.addNewTool(SDLK_F7, new RHCopyTool(&sel));
    sel.addNewTool(SDLK_F8, new RHPasteTool(true));
    sel.addNewTool(SDLK_F9, new RHPasteTool(false));
    painter.setZoom(2);

    setLayout(UIHBoxLayout::instance5);
    add(g1);
    add(g2);
    add(g3);

    g1.add(sel);
    g1.add(h.add(new RHZoomSelector(painter, 4)));
    g1.add(h.add(new UISpacer()));

    painter_widget = &painter;
    g2.add(g21);
    g2.add(painter);

    charname_widget = &h.add(new RHCharacterName());
    g21.add(h.add(new UIButton("<", SDLK_PAGEUP)));
    g21.add(*charname_widget);
    g21.add(h.add(new UIButton("G", 'g')));
    g21.add(h.add(new UIButton(">", SDLK_PAGEDOWN)));

    g3.add(h.add(new RHFontColorSelector(painter)));
    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(0, _("Close"), SDLK_ESCAPE)).addStop());

    pack();

    /* Figure out horizontal aux lines: common heights */
    // regular character top
    int n;
    if (const GfxBitmapGlyph* g = font->getGlyph('B'))
        if ((n = findTop(*g)) >= 0)
            painter.addAuxLine(n, false);

    // Lower-case top
    if (const GfxBitmapGlyph* g = font->getGlyph('o'))
        if ((n = findTop(*g)) >= 0)
            painter.addAuxLine(n, false);

    // Bottom
    if (const GfxBitmapGlyph* g = font->getGlyph('o'))
        if ((n = findBottom(*g)) >= 0)
            painter.addAuxLine(n, false);
    
    // Descender
    if (const GfxBitmapGlyph* g = font->getGlyph('g'))
        if ((n = findBottom(*g)) >= 0)
            painter.addAuxLine(n, false);

    /* Figure out vertical aux lines: common widths. This is to
       upport making fixed-width fonts */
    enum { MAXWIDTHS = 3 };
    int nwidths = 0;
    int widths[MAXWIDTHS];
    bool ok = true;
    for (uint32_t i = 0, e = font->getCurrentCharacterLimit(); i < e; ++i) {
        if (const GfxBitmapGlyph* g = font->getGlyph(i)) {
            bool found = false;
            for (int j = 0; j < nwidths; ++j)
                if (widths[j] == g->getWidth())
                    found = true;
            if (!found) {
                if (nwidths < MAXWIDTHS) {
                    widths[nwidths++] = g->getWidth();
                } else {
                    ok = false;
                    break;
                }
            }
        }
    }
    if (ok) {
        for (int i = 0; i < nwidths; ++i)
            painter.addAuxLine(widths[i], true);
    }
}

void
RHFontEditorWindow::setCharacter(uint32_t ch)
{
    if (ch != current_character) {
        font->addNewGlyph(current_character, pixmapToGlyph(painter_widget->getPixmap()));
        current_character = ch;
        charname_widget->setCharacter(ch);
        painter_widget->setPixmap(glyphToPixmap(*font, ch));
    }
}

bool
RHFontEditorWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        uint32_t n;
        switch (event.key.code) {
         case SDLK_PAGEUP:
            setSelectedWidget();
            if (current_character > 0) {
                setCharacter(current_character - 1);
            }
            return true;
         case SDLK_PAGEDOWN:
            setSelectedWidget();
            if (current_character < 0xFFFF) {
                setCharacter(current_character + 1);
            }
            return true;
         case SDLK_PAGEUP + ss_Ctrl:
            setSelectedWidget();
            if (current_character > 0) {
                setCharacter(getNextExistingCharacter(*font, current_character, -1));
            }
            return true;
         case SDLK_PAGEDOWN + ss_Ctrl:
            setSelectedWidget();
            if (current_character < 0xFFFF) {
                setCharacter(getNextExistingCharacter(*font, current_character, +1));
            }
            return true;
         case '+':
            setSelectedWidget();
            painter_widget->setSize(painter_widget->getSize().movedBy(1, 0));
            return true;
         case '-':
            setSelectedWidget();
            if (painter_widget->getSize().x != 0) {
                painter_widget->setSize(painter_widget->getSize().movedBy(-1, 0));
            }
            return true;
         case SDLK_DELETE:
            setSelectedWidget();
            if (painter_widget->getSize().x != 0 && UIMessageBox(_("Delete this character?"), _("Delete Character")).doYesNoDialog()) {
                painter_widget->setSize(GfxPoint(0, painter_widget->getSize().y));
            }
            return true;
         case 'a':
            setSelectedWidget();
            if (!RHEditorCharacterGenerator(*painter_widget, *font).generate(current_character)) {
                if (!synthesizeCharacter(*painter_widget, *font, current_character)) {
                    messageBox(_("I do not know how to generate this character."), _("Auto-Generate Character"));
                }
            }
            return true;
         case 'C':
         case 'c':
            setSelectedWidget();
            storeCurrentCharacter();
            if (pickCharacter(_("Copy From Character"), n, event.key.code == 'C')) {
                painter_widget->setPixmap(glyphToPixmap(*font, n));
            }
            return true;
         case 'e':
            setSelectedWidget();
            storeCurrentCharacter();
            if (doFontCoverageDialog(*font, n)) {
                setCharacter(n);
            }
            return true;
         case 'E':
            setSelectedWidget();
            doCharacterCodeInfo(current_character);
            return true;
         case 'G':
         case 'g':
            setSelectedWidget();
            storeCurrentCharacter();
            if (pickCharacter(_("Go To Character"), n, event.key.code == 'G')) {
                setCharacter(n);
            }
            return true;
         case 'm':
            setSelectedWidget();
            storeCurrentCharacter();
            if (Ptr<GfxBitmapFont> newFont = doVerticalAlignDialog(*font)) {
                font = newFont;
                painter_widget->setPixmap(glyphToPixmap(*font, current_character));
            }
            return true;
         case 'O':
         case 'o':
            setSelectedWidget();
            storeCurrentCharacter();
            if (pickCharacter(_("Overlay Character"), n, event.key.code == 'O')) {
                if (const GfxBitmapGlyph* g = font->getGlyph(n)) {
                    // FIXME: it would make sense to use "maximum" logic, i.e. never
                    // draw a FC_HALF pixel above a FC_WHITE pixel.
                    g->drawColored(*painter_widget->getPixmap(), 0, 0, FC_WHITE, FC_HALF);
                    painter_widget->drawWidget();
                }
            }
            return true;
         case 'r':
            setSelectedWidget();
            {
                StringList sl;
                for (std::size_t i = 0; i < CharacterSet::getNumCharsets(); ++i)
                    sl.add(i, format(_("%s to Unicode"), CharacterSet::getCharset(i).getName()));
                UIStandardListbox lb(0, -1, getTextWidth(*fonts[FONT_NORMAL], sl), sl);
                if (lb.doStandardListbox(_("Change Encoding"))) {
                    storeCurrentCharacter();
                    changeToUnicode(*font, CharacterSet::getCharset(lb.getKey()));
                    painter_widget->setPixmap(glyphToPixmap(*font, current_character));
                }
            }
            return true;
         case 's':
            setSelectedWidget();
            {
                int of = showCharacterInSystemFonts(current_character);
                if (of >= 0) {
                    if (GfxBitmapFont* bmfont = dynamic_cast<GfxBitmapFont*>(fonts[of].ptr())) {
                        if (const GfxBitmapGlyph* g = bmfont->getGlyph(current_character)) {
                            // Glyph exists in other font. Make sure this glyph
                            // is wide enough, then copy it
                            GfxPoint pt = painter_widget->getSize();
                            if (pt.x < g->getWidth()) {
                                pt.x = g->getWidth();
                                painter_widget->setSize(pt);
                            }
                            g->drawColored(*painter_widget->getPixmap(), 0, 0, FC_WHITE, FC_HALF);
                            painter_widget->drawWidget();
                        }
                    }
                }
            }
            return true;
         case 'u':
            setSelectedWidget();
            if (UIMessageBox(_("Undo changes to this character and restore previous content?"),
                             _("Font Editor")).doYesNoDialog())
            {
                painter_widget->setPixmap(glyphToPixmap(*font, current_character));
            }
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

/** Pick a character.
    \param title  [in] Prompt
    \param result [out] Result will be stored here
    \param withGrid [in] true to use grid, false to use list
    \return true iff user chose OK */
bool
RHFontEditorWindow::pickCharacter(string_t title, uint32_t& result, bool withGrid)
{
    result = current_character;
    if (withGrid) {
        return RHCharacterGridDialog::doDialog(*font, title, result);
    } else {
        return RHPickCharacterDialog::doDialog(title, result);
    }
}

/** Store current character in font. Use this when the editor ends to
    store the last changes.*/
void
RHFontEditorWindow::storeCurrentCharacter()
{
    font->addNewGlyph(current_character, pixmapToGlyph(painter_widget->getPixmap()));
}


/************************* RHPickCharacterDialog *************************/

/** Constructor.
    \param title Window title */
RHPickCharacterDialog::RHPickCharacterDialog(string_t title)
    : UIWindow(title, BLUE_WINDOW, COLOR_GRAY),
      list_widget(0),
      input_widget(0)
{
    init();
} 
    
void
RHPickCharacterDialog::init()
{
    // VBox
    //   UIInputLine
    //   HBox
    //     RHCharacterNameList
    //     UIScrollbar
    //   HBox
    //     UISpacer
    //     UIButton "OK"
    //     UIButton "Cancel"

    setLayout(UIVBoxLayout::instance5);
    UIFrameGroup& g2 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIGroup& g3 = h.add(new UIGroup(UIHBoxLayout::instance5));

    input_widget = &h.add(new UIInputLine(1000, 20, 0, UIInputLine::ilf_Framed, 1000));
    input_widget->sig_changed.add(this, &RHPickCharacterDialog::onType);
    add(*input_widget);
    add(g2);
    add(g3);

    list_widget = &h.add(new RHCharacterNameList(1001, list));
    sb_widget = &h.add(new UIScrollbar(*list_widget));
    g2.add(*list_widget);
    g2.add(*sb_widget);

    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(cm_Enter,  _("OK"), SDLK_RETURN)).addStop());
    g3.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());

    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab | UIFocusIterator::fi_Wrap)));

    pack();
    onType();
}

/** Handle changes in the input line. Updates the filtered list. */
void
RHPickCharacterDialog::onType()
{
    matchCharacterNames(input_widget->getText(), 0xFFFF, list);
    list_widget->update();
    sb_widget->drawWidget();
    // sb_widget->sync(); FIXME -- doesn't work currently
    setStateOf(cm_Enter, st_Disabled, list.empty());
}

/** Invoke this dialog.
    \param title [in] Window title
    \param result [out] Selected character code
    \return true iff dialog was successful, false if canceled */
bool
RHPickCharacterDialog::doDialog(string_t title, uint32_t& result)
{
    RHPickCharacterDialog dlg(title);
    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg) == cm_Enter && !dlg.list.empty()) {
        result = dlg.list[dlg.list_widget->getCurrent()];
        return true;
    } else {
        return false;
    }
}

/************************* RHCharacterGridDialog *************************/


RHCharacterGridDialog::RHCharacterGridDialog(const GfxBitmapFont& font, string_t title)
    : UIWindow(title, BLUE_WINDOW, COLOR_BLACK)
{
    init(font);
}

void
RHCharacterGridDialog::init(const GfxBitmapFont& font)
{
    // VBox
    //   RHCharacterGrid
    //   RHCharacterName
    //   HBox
    //     UISpacer
    //     "OK"
    //     "Cancel"

    // Main widgets
    grid_widget = &h.add(new RHCharacterGrid(font));
    grid_widget->sig_moved.add(this, &RHCharacterGridDialog::onMove);
    charname_widget = &h.add(new RHCharacterName());

    // Buttons
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UIButton(2, _("Cancel"), SDLK_ESCAPE)).addStop());

    // Build it
    setLayout(UIVBoxLayout::instance5);
    add(*grid_widget);
    add(*charname_widget);
    add(g);
    pack();

    onMove();
}

void
RHCharacterGridDialog::onMove()
{
    charname_widget->setCharacter(grid_widget->getCurrentCharacter());
}

bool
RHCharacterGridDialog::doDialog(const GfxBitmapFont& font, string_t title, uint32_t& result)
{
    RHCharacterGridDialog dlg(font, title);
    ui_root->centerWidget(dlg);
    dlg.grid_widget->setCurrentCharacter(result);
    if (ui_root->runModal(dlg) == 1) {
        result = dlg.grid_widget->getCurrentCharacter();
        return true;
    } else {
        return false;
    }
}

/************************ RHSystemCharacterWidget ************************/

RHSystemCharacterWidget::RHSystemCharacterWidget(uint32_t ch)
    : ch(ch)
{ }

void
RHSystemCharacterWidget::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.setColor(UIColor::tc_Static);

    string_t str;
    appendUtf8(str, ch);
    int y = getExtent().y;
    int x = getExtent().x;
    int x2 = x + 2*font_ems[FONT_NORMAL];
    for (int i = 0; i < MIN_FONTS; ++i) {
        ctx.useFont(*fonts[FONT_NORMAL]);
        outText(ctx, x, y, format("%d:", i));
        ctx.useFont(*fonts[i]);
        outText(ctx, x2, y, str);

        int dy = fonts[i]->getTextHeight(str);
        if (dy < font_heights[FONT_NORMAL])
            dy = font_heights[FONT_NORMAL];
        y += dy;
    }

    can.update(getExtent());
}

bool
RHSystemCharacterWidget::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code >= '0' && event.key.code < '0' + MIN_FONTS) {
        setSelectedWidget();
        stop(event.key.code - '0');
        return true;
    } else {
        return UIWidget::handleEvent(event, second_pass);
    }
}

void
RHSystemCharacterWidget::getLayoutInfo(LayoutInfo& info)
{
    string_t str;
    appendUtf8(str, ch);

    GfxPoint pt(0, 0);
    for (int i = 0; i < MIN_FONTS; ++i) {
        int dx = fonts[i]->getTextWidth(str);
        int dy = fonts[i]->getTextHeight(str);

        if (dx > pt.x)
            pt.x = dx;

        if (dy < font_heights[FONT_NORMAL])
            dy = font_heights[FONT_NORMAL];
        pt.y += dy;
    }
    pt.x += 2*font_ems[FONT_NORMAL];
    info.min_size = info.pref_size = pt;
    info.flags = 0;
}

/*********************** RHEditorCharacterGenerator **********************/

static bool
checkOverlap(GfxPixmap& pixmap, const GfxBitmapGlyph& glyph, int dx, int dy)
{
    for (int y = 0; y < glyph.getHeight(); ++y) {
        for (int x = 0; x < glyph.getWidth(); ++x) {
            /* This uses that getPixel for an out-of-range pixel returns 0 */
            if (glyph.get(x, y) && pixmap.getPixel(GfxPoint(x+dx, y+dy)) != FC_BLACK) {
                return true;
            }
        }
    }
    return false;
}

RHEditorCharacterGenerator::RHEditorCharacterGenerator(RHPainter& painter, GfxBitmapFont& font)
    : painter(painter), font(font)
{ }

bool
RHEditorCharacterGenerator::check(uint32_t chars[], uint32_t nchars)
{
    // Check whether we have the required characters, and generate prompt on the fly
    string_t msg = _("Generate this character from the following?");
    for (uint32_t i = 0; i < nchars; ++i) {
        if (!font.getGlyph(chars[i]))
            return false;
        msg += "\n";
        msg += getCharacterName(chars[i]);
    }

    // Ask user
    if (!UIMessageBox(msg, _("Auto-Generate Character")).doYesNoDialog())
        return false;

    // Figure out width
    int width = 0;
    for (uint32_t i = 0; i < nchars; ++i) {
        int thiswidth = font.getGlyph(chars[i])->getWidth();
        if (thiswidth > width)
            width = thiswidth;
    }

    // Build result
    Ptr<GfxPixmap> result = new GfxPixmap(width, painter.getPixmap()->getHeight(), *painter.getPixmap());
    for (uint32_t i = 0; i < nchars; ++i) {
        const GfxBitmapGlyph* g = font.getGlyph(chars[i]);
        int dx = (width - g->getWidth())/2;
        int dy = 0;
        if (i != 0) {
            dy = -g->getHeight();
            while (!checkOverlap(*result, *g, dx, dy) && dy < g->getHeight())
                ++dy;
            dy -= 2;
        }
        
        g->drawColored(*result, dx, dy, FC_WHITE, FC_HALF);
    }

    painter.setPixmap(result);
    return true;
}

/**************************** RHSaveFontDialog ***************************/

RHSaveFontDialog::RHSaveFontDialog()
    : UIWindow(_("Save Font"), BLUE_WINDOW, COLOR_GRAY),
      encoding(0),
      file_name(1000, 20, ss_Alt + 'n', UIInputLine::ilf_Framed, 1000),
      font_name(256,  20, ss_Alt + 'f', UIInputLine::ilf_Framed, 1001)
{
    init();
}

RHSaveFontDialog::~RHSaveFontDialog()
{ }

void
RHSaveFontDialog::init()
{
    // VBox
    //   HBox
    //     "File Name"
    //     UIInputLine 1000
    //   HBox
    //     "Font Name"
    //     UIInputLine 1001
    //   HBox
    //     "Encoding"
    //     VBox        1002
    //       UIRadioButton^n
    //   HBox
    //     UISpacer
    //     UIButton "OK"
    //     UIButton "Cancel"

    setLayout(UIVBoxLayout::instance5);
    UIGroup& g1  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g2  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g3  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g32 = h.add(new UIGroup(UIHBoxLayout::instance5, 1002));
    UIGroup& g4  = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIStaticText(_("File Name:"), UIStaticText::StaticTitle)));
    g1.add(file_name);

    g2.add(h.add(new UIStaticText(_("Font Name:"), UIStaticText::StaticTitle)));
    g2.add(font_name);

    g3.add(h.add(new UIStaticText(_("Encoding:"), UIStaticText::StaticTitle)));
    g3.add(g32);

    g32.add(h.add(new UIRadioButton(2000, _("System"),            ss_Alt + 's', encoding, 0)));
    g32.add(h.add(new UIRadioButton(2001, _("Cyrillic (CP866)"),  ss_Alt + 'c', encoding, 1)));
    g32.add(h.add(new UIRadioButton(2002, _("Unicode"),           ss_Alt + 'u', encoding, 2)));

    g4.add(h.add(new UISpacer()));
    g4.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g4.add(h.add(new UIButton(2, _("Cancel"), SDLK_ESCAPE)).addStop());

    add(g1);
    add(g2);
    add(g3);
    add(g4);
    add(h.add(new UIFocusIterator(1000, 1003, UIFocusIterator::fi_Tab | UIFocusIterator::fi_Vert)));
    add(h.add(new UIFocusIterator(2000, 2002, UIFocusIterator::fi_Horiz)));
    pack();
}

/****************************** RHFontEditor *****************************/

RHFontEditor::RHFontEditor(Ptr<GfxBitmapFont> font, string_t file_name)
    : font(font),
      file_name(file_name),
      dirty(false)
{ }

RHFontEditor::~RHFontEditor()
{ }

string_t
RHFontEditor::getName()
{
    string_t name;
    if (file_name.empty())
        name = _("Unnamed");
    else
        name = Directory::getFileName(file_name);
    name += format(_(" (font, height %d pixels)"), font->getTextHeight("Tp"));
    return name;
}

void
RHFontEditor::edit()
{
    RHFontEditorWindow win(font, Directory::getFileName(file_name));
    win.setExtent(ui_root->getExtent());
    ui_root->runModal(win);
    win.storeCurrentCharacter();
    font = win.getFont();
    dirty = true;
}

void
RHFontEditor::save()
{
    // Dialog
    RHSaveFontDialog dlg;
    dlg.setFileName(file_name);
    dlg.setEncoding(0);

    ui_root->centerWidget(dlg);
    if (ui_root->runModal(dlg) == 1) {
        try {
            saveFont(dlg.getFileName(), dlg.getFontName(), dlg.getEncoding() * 256);
            dirty = false;
        }
        catch (FileProblemException& e) {
            messageBox(format(_("Unable to save file %s: %s"), e.getFileName(), e.what()), _("Error"));
        }
        catch (std::exception& e) {
            messageBox(format(_("Unable to save file %s: %s"), dlg.getFileName(), e.what()), _("Error"));
        }
    }
}

static void
saveFontHeader(Stream& s,
               uint32_t& font_address,
               const string_t& comment,
               int flags,
               GfxBitmapFont& font,
               std::vector<uint32_t>& character_ids,
               std::vector<uint32_t>& character_addresses)
{
    // Main header:
    //  2 bytes signature
    //  2 bytes number of fonts (=1)
    //  4 bytes font address
    //  4 bytes flags
    //  1 byte font name (comment) length
    //  font name

    char header[13];
    header[0] = 'F';
    header[1] = 'N';
    header[2] = 1;
    header[3] = 0;
    storeUint32(header+4, font_address);
    storeUint32(header+8, flags);
    header[12] = comment.size() > 255 ? char(255) : char(comment.size());
    s.writeT(header, sizeof(header));

    // Font name
    s.writeT(comment.data(), uint8_t(header[12]));

    // Font data. font_address points here.
    font_address = s.getPos();

    // 2 bytes type (=3)
    // 2 bytes height
    // 2 bytes number of characters
    storeUint16(header,   3);
    storeUint16(header+2, font.getTextHeight("Tp"));
    storeUint16(header+4, character_ids.size());
    s.writeT(header, 6);

    // Character headers
    for (uint32_t i = 0; i < character_ids.size(); ++i) {
        // 4 bytes address
        // 2 bytes id
        // 2 bytes width
        const GfxBitmapGlyph* g = font.getGlyph(character_ids[i]);
        storeUint32(header, character_addresses[i]);
        storeUint16(header+4, character_ids[i]);
        storeUint16(header+6, g ? g->getWidth() : 0);
        s.writeT(header, 8);
    }
}

void
RHFontEditor::saveFont(string_t name, string_t comment, int flags)
{
    // Build list of characters
    std::vector<uint32_t> character_ids;
    uint32_t max = font->getCurrentCharacterLimit();
    if (max > 0xFFFF)
        max = 0xFFFF;
    if (max < 0x100)
        max = 0x100;
    for (uint32_t i = 0; i < max; ++i) {
        if (i < 0x100 || font->getGlyph(i) != 0) {
            character_ids.push_back(i);
        }
    }

    // Open file
    Ptr<Stream> s = Stream::openFile(name, Stream::C_CREATE);

    // Prepare header
    std::vector<uint32_t> character_addresses(character_ids.size());
    uint32_t font_address = 0;
    saveFontHeader(*s, font_address, comment, flags, *font, character_ids, character_addresses);

    // Save characters
    for (uint32_t i = 0; i < character_ids.size(); ++i) {
        const GfxBitmapGlyph* g = font->getGlyph(character_ids[i]);
        character_addresses[i] = s->getPos();
        if (g != 0) {
            // Glyph exists
            const std::vector<uint8_t>& data = g->getData();
            const std::vector<uint16_t>& aadata = g->getAAData();
            s->writeT((const char*)&data[0], data.size());

            uint16_t naa = aadata.size()/2;
            std::vector<char> a(naa*4 + 2);
            storeUint16(&a[0], naa);
            for (uint16_t i = 0; i < naa; ++i) {
                storeUint16(&a[4*i+2], aadata[2*i]);
                storeUint16(&a[4*i+4], aadata[2*i+1]);
            }
            s->writeT(&a[0], a.size());
        } else {
            // Glyph does not exist. Write just a blank number-of-AA-hints entry
            char tmp[2];
            tmp[0] = tmp[1] = 0;
            s->writeT(tmp, 2);
        }
    }

    // Update header
    s->seek(0);
    saveFontHeader(*s, font_address, comment, flags, *font, character_ids, character_addresses);
}

bool
RHFontEditor::hasUnsavedChanges()
{
    return dirty;
}
