/**
  *  \file util/filepars.cc
  *  \brief File Parser
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <cstring>
#include "util/filepars.h"
#include "cpluslib/string.h"
#include "util/misc.h"
#include "io/textfile.h"
#include "io/streams.h"

/** \class FileParser
    \brief Abstract File Parser Class

    This class provides the logic to read text files. It will split
    the file into lines, wield out empty and comment lines, and call
    process() or unparsed() for each line. Derived classes will
    override these methods to actually do some parsing.

    Comments are started by using a ';' as first non-blank character.

    In addition, this provides methods to read files from game or root
    directory. */

/** Constructor. Does nothing special. */
FileParser::FileParser(const char* comment_chars)
    : name("<unnamed>"), comment_chars(comment_chars),
      utf8(false), charset()
{ }

/** Callback: parse a text line. This function is called for each line
    which is neither empty nor a comment.
    \param line_id     line number
    \param line        line text */
void
FileParser::process(int /*line_id*/, string_t /*line*/)
{ }

/** Callback: parse commetn line. This function is called for each
    line which is either empty or a comment. */
void
FileParser::unparsed(int /*line_id*/, string_t /*line*/)
{ }

/** Parse a file.
    \param in stream to read from */
void
FileParser::parseFile(Stream& s)
{
    using namespace std; // strchr

    TextFile tf(s);
    if (utf8)
        tf.setUtf8();
    else
        tf.setCharacterSet(charset);
    name = s.getName();

    string_t line;
    while (tf.getLine(line)) {
        string_t::size_type n = 0;
        while (n < line.length() && isspace(uint8(line[n])))
            ++n;
        if (n >= line.length() || (comment_chars != 0 && strchr(comment_chars, line[n])))
            unparsed(tf.getLineNr(), line);
        else
            process(tf.getLineNr(), line);
    }
}

/** Trim comments. If the line contains a comment, remove it. */
void
FileParser::trimComments(string_t& line)
{
    using namespace std;
    if (comment_chars != 0) {
        string_t::size_type n = line.find_first_of(comment_chars);
        if (n != string_t::npos) {
            while (n > 0 && isspace(line[n-1]))
                --n;
            line.erase(n);
        }
    }
}

/** Set single-byte character set. */
void
FileParser::setCharacterSet(CharacterSet cs)
{
    this->charset = cs;
    this->utf8 = false;
}

/** Set UTF-8. */
void
FileParser::setUtf8()
{
    this->utf8 = true;
}
